#include "clock8e.inc"
// #include "./clock72.inc"
// #include "./clock216.inc"
#include <libopencm3/stm32/rcc.h>
#include <libopencm3/stm32/gpio.h>
#include <libopencm3/cm3/scb.h>
#include <libopencm3/stm32/usart.h>
#include <libopencm3/cm3/nvic.h>
#include <libopencm3/cm3/systick.h>
#include "stmf1usart.h"
//#include "../f4/cm4scb.h"

/* Testing time taken by an interrupt with output via UART.
   Pins:
   PA9    USART1_TX
   PA10   USART1_RX
*/

static void wait_usart_quiet(void);
uint32_t millis(void);

static int brrs[] = {
   69, // 8000/115.2
   35, // 4000/115.2
   17, // 2000/115.2
};

void
set_apb2_div(int k) {
    int k1 = (k == 0)?0:(3 + k);
    wait_usart_quiet();
    rcc->cfgr = (rcc->cfgr & ~RCC_CFGR_PPRE2) |
                (k1 << RCC_CFGR_PPRE2_SHIFT);
    /* Set new baudrate */
    usart1->brr = brrs[k];
}

void
set_flash_lat(int k) {
    FLASH_ACR = (FLASH_ACR & ~0x7) | k;    
}

void
wait_usart_quiet(void) {
    while(!(usart1->sr & USART_SR_TC)) {
        ;
    }
}

void
xmit_byte(uint8_t c)
{
    while(!(usart1->sr & USART_SR_TXE))
        ;
    usart1->dr = c;
}

void pend_sv_handler(void) {
    asm("nop");
    asm("nop");
}

// #define systick_reload (216000 - 1)
// #define systick_reload (72000 - 1)
#define systick_reload (8000 - 1)

volatile uint32_t systick_milliseconds = 0;

void sys_tick_handler(void)
{
    // 27 by tylko podstawiać, 28 by wywołać PendSv
    // scb->icsr = (1<<27);
    systick_milliseconds++;
}

void sys_tick_init(void)
{
    STK_RVR = systick_reload;
    STK_CVR = 0;
    STK_CSR = STK_CSR_CLKSOURCE_AHB | STK_CSR_TICKINT | STK_CSR_ENABLE;
}

uint32_t millis(void)
{
    return systick_milliseconds;
}

uint32_t
test_delay(void (*fn)(uint32_t, uint32_t, uint32_t, uint32_t),
           uint32_t cnt, uint32_t addr) {
    uint32_t t1 = STK_CVR;
    (*fn)(cnt, addr, 0, 1);
    return t1 - STK_CVR;
}

void print_unum(uint32_t num) {
    char buff[12];
    int i = 11;
    buff[i]=0;
    while(1) {
        uint32_t nnum = num/10;
        i--;
        buff[i] = '0' + num - 10*nnum;
        if(num<10) {
                break;
        }
        num = nnum;
    }
    while(i < 11) {
        xmit_byte(buff[i]);
        i++;
    }
}

void
print_string(char * s) {
    char c;
    while ((c = *s++)) {
        xmit_byte(c);
    }
}

uint32_t curr_time = 0;

struct tst_pos {
   void (*fn)(uint32_t, uint32_t, uint32_t, uint32_t);
   uint32_t cnt;
   uint32_t addr;
};

uint32_t
do_tst(struct tst_pos * p0, int n, int k) {
    struct tst_pos * p = p0 + n;
    return test_delay(p->fn, p->cnt+k, p->addr);
}

extern void counted_delay(uint32_t t, uint32_t, uint32_t, uint32_t);
extern void pin_test1(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void pin_test2(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void pin_test3(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void pin_test4(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void pin_test11(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void pin_test12(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void pin_test5(uint32_t c, uint32_t a, uint32_t v0, uint32_t v1);
extern void del_tst2(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void del_tst3(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void del_tst4(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void del_tst5(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void del_tst6(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void del_tst7(uint32_t c, uint32_t, uint32_t, uint32_t);
extern void del_tst8(uint32_t c, uint32_t, uint32_t, uint32_t);

uint32_t vvvv;

struct tst_pos tst_tab[] = {
    {pin_test1, 100, 0},
    {pin_test2, 100, 0},
    {pin_test5, 100, 0x42210198},
    {pin_test5, 100, 0x4001080c},
    {pin_test3, 100, 0},
    {pin_test4, 100, 0},
    {pin_test5, 100, 0x42400180},
    {pin_test5, 100, 0x4002000C},
    {pin_test11, 100, 0},
    {pin_test12, 100, 0},
    {pin_test5, 100, 0x22080000 /* bitband 0x20004000 */},
    {pin_test5, 100, (uint32_t)(&vvvv)},
    {counted_delay, 513, 0},
    {del_tst2, 200, 0},
    {del_tst3, 200, 0},
    {del_tst4, 100, 0},
    {del_tst5, 50, 0},
    {del_tst6, 20, 0},
    {del_tst7, 20, 0},
    {del_tst8, 20, 0},
};

#define N_TST (sizeof(tst_tab)/sizeof(struct tst_pos))

char quiet_flag = 5;

int cur_lat = 0;
int apb_cur_step = 0;
int tst_cur_n = 0;
int tst_cur_k = 0;


#define N_K 12
uint16_t vals[N_K];

void
print_line(int k) {
    int i;
    uint16_t val0 = vals[0];
    if (k < 10) {
        xmit_byte(' ');
    }
    print_unum(k);
    xmit_byte(' ');
    xmit_byte(' ');
    if (val0 < 1000) {
        xmit_byte(' ');
    }
    if (val0 < 100) {
        xmit_byte(' ');
    }
    print_unum(val0);
    for(i = 1; i < N_K; i++) {
        uint16_t val1 = vals[i];
        val0 = val1 - val0;
        xmit_byte(' ');
        if(val0 < 100) {
            xmit_byte(' ');
        }
        if(val0 < 10) {
            xmit_byte(' ');
        }
        print_unum(val0);
        val0 = val1;
    }
    xmit_byte('\n');
}

void loop()
{
    uint32_t n_time = millis();
    if (curr_time != n_time) {
        curr_time = n_time;
    } else {
        return;
    }
    if (quiet_flag == 0) {
        if (cur_lat > 5) {
            return;
        }
        quiet_flag = 5;
        uint32_t val = do_tst(tst_tab, tst_cur_n, tst_cur_k);
        vals[tst_cur_k] = val;
        tst_cur_k++;
        if (tst_cur_k < N_K) {
            return;
        }
        tst_cur_k = 0;
        print_line(tst_cur_n);
        tst_cur_n++;
        if ((tst_cur_n == N_TST)|| ((apb_cur_step > 0) && (tst_cur_n == 4))) {
            tst_cur_n = 0;
            apb_cur_step++;
            if (apb_cur_step == 3) {
                apb_cur_step = 0;
            }
            set_apb2_div(apb_cur_step);
            if (apb_cur_step > 0) {
                return;
            }
        } else {
            return;
        }
        cur_lat++;
        xmit_byte('\n');
        if (cur_lat <= 5) {
            set_flash_lat(cur_lat);
            print_string("Latency ");
            print_unum(cur_lat);
            xmit_byte('\n');
        }
    } else {
        quiet_flag--;
    }
}


__attribute__((optimize(0))) void configure_vector(void)
{
    // This takes _data_loadaddr from the linker file
    // And evaluates wether the address is a ROM or RAM 
    // address. This does not work properly with optimisation enabled
    // therefore optimisation is disabled for this function.
    extern volatile uint32_t _data_loadaddr;
    uint32_t ramrom = (uint32_t)&_data_loadaddr;
    bool isram = (ramrom&0xF0000000) == 0x20000000;

    if (isram)
        SCB_VTOR = 0x20000000;
}


int main(void)
{
    // Configure the vector table if appropiate.
    configure_vector();

    /* Configure the clock */
    clock_hse8();
    // clock_72mhz_from_hse8();
    // clock_216mhz_from_hse8();

    /* Activate UART1 GPIO lines */
    rcc->apb2enr |= RCC_APB2ENR_USART1EN | RCC_APB2ENR_IOPAEN
                   | RCC_APB2ENR_AFIOEN;
    rcc->ahbenr |= RCC_AHBENR_DMA1EN;
    gpio_set_mode(GPIOA, GPIO_MODE_INPUT,
                  GPIO_CNF_INPUT_FLOAT, GPIO10);
    // GPIO_MODE_OUTPUT_2_MHZ is too weak above 460800
    // GPIO_MODE_OUTPUT_10_MHZ is OK for 921600
    gpio_set_mode(GPIOA, GPIO_MODE_OUTPUT_50_MHZ,
                  GPIO_CNF_OUTPUT_ALTFN_PUSHPULL, GPIO9);

    /* Configure UART */
    usart1->brr = 69 ; // 8000/115.2 
    // usart1->brr = 625 ; // 72000/115.2
    // usart1->brr = 1875 ; // 216000/115.2
    usart1->cr1 = USART_CR1_TE | USART_CR1_UE ;

    sys_tick_init();

    /* Make sure interrupts are enabled */
    asm("cpsie i");

    print_string("Latency 0\n");

    while(1) {
        asm("wfi");
        loop();
    }
}
